class HTTPException:
    def __init__(self, response, error_string: str = None):
        self.status = response.status
        self.message = error_string or f'The server responded with a status of {self.status}'
        self.response = response

        super().__init__(self.message)
    
    def __repr__(self) -> str:
        raise f"<HTTPException status={self.status} response={self.response!r}>"
    
    def __int__(self) -> int:
        return self.status


from .rest import HTTPClient
from gc import collect
from urllib.parse import unquote, quote
from re import search

class Client:
    __slots__ = ('http',)

    def __init__(self, session=None, format: str = 'C', locale: str = 'en-US', max_cache_size: int = 15):
        if format.upper() not in ('C', 'F'):
            raise TypeError('Invalid format.')
        
        self.http = HTTPClient(session, max_cache_size, format, locale)

    @property
    def format(self) -> str:
        return search('&weadegreetype=([^&]+)', self.http._query_params)[0][15:]

    @format.setter
    def format(self, value: str) -> None:
        if value.upper() not in ('C', 'F'):
            raise TypeError('Invalid format.')
        self.http._query_params = self.http._query_params.replace(f'&weadegreetype={self.format}', f'&weadegreetype={value}')

    @property
    def locale(self) -> str:
        return unquote(search('&culture=([^&]+)', self.http._query_params))[0][9:]

    @locale.setter
    def locale(self, value: str) -> None:
        self.http._query_params = self.http._query_params.replace(f'&culture={self.locale}', f'&culture={quote(value)}')

    async def find(self, location: str) -> "Weather":
        """ Finds a weather forecast from a location. """
        if (not location) or (not isinstance(location, str)):
            raise TypeError('location must be a string.')
        return await self.http.request(location)

    @property
    def cache(self) -> "Cache":
        """ Returns the cache. """
        if self.closed:
            return
        return self.http.cache
    
    @cache.setter
    def cache(self, new_cache: dict) -> None:
        """ Modifies the cache. """
        if self.closed:
            return
        elif new_cache is None:
            return self.cache.clear()
        elif not isinstance(new_cache, dict):
            raise TypeError('Invalid type for setting the client\'s cache.')

        new_is_empty = not len(new_cache.keys())
        for k, v in new_cache.items():
            self.http.cache[k] = v
        
        if new_is_empty:
            collect()

    @property
    def closed(self) -> bool:
        """ Returns if the client is closed or not. """
        return self.http.closed

    async def close(self) -> None:
        """ Closes the wrapper. """
        if not self.closed:
            await self.http.close()

    def __repr__(self) -> str:
        return f"<WeatherClient closed={self.closed}>"


from asyncio import sleep
from xmltodict import parse
from aiohttp import ClientSession
from urllib.parse import quote
from .cache import Cache
from .response import Weather
from .exceptions import HTTPException

class HTTPClient:
    __slots__ = ('_session', '_query_params', 'cache')

    def __init__(self, max_cache_size: int, format: str, locale: str, session: "ClientSession" = None):
        self._session = session if isinstance(session, ClientSession) and (not session.closed) else ClientSession()
        self._query_params = f"?src=outlook&culture={quote(locale)}&weadegreetype={format}&weasearchstr=" + "{}"
        self.cache = Cache(max_cache_size)

    async def request(self, location: str) -> dict:
        url = "https://weather.service.msn.com/find.aspx" + self._query_params.format(quote(location))

        if url not in self.cache:
            resp = await self._session.get(url)
            self.cache[url] = await resp.text()
        
        parsed = parse(self.cache[url])
        if (error_message := parsed.get('string')):
            raise HTTPException(response, error_message['#text'])
        elif resp.status >= 400:
            raise HTTPException(response)
        
        resp.close()
        try:
            return Weather(parsed['weatherdata']['weather'][0])
        except KeyError:
            return Weather(parsed['weatherdata']['weather'])
    
    @property
    def closed(self) -> bool:
        """ Returns if the HTTPClient is closed. """
        return self._session.closed and not bool(self.cache)

    def __repr__(self) -> str:
        return f"<HTTPClient closed={self.closed}>"

    async def close(self):
        if not self._session.closed:
            await self._session.close()
        
        self.cache.clear()


from halo import Halo
halo_ = Halo(spinner='lines', text='please wait...', color='white').start()
client_ = None

from python_weather import Client, HTTPException
from typing import Optional
from asyncio import get_event_loop
from sys import argv, exit

flags = [
    ('help', 'Shows this help screen.'),
    ('unit C | F | METRIC | IMPERIAL', 'Sets the measuring unit to metric (default) or imperial.'),
    ('locale <LANG>', 'Sets the default locale. Defaults to en-US.'),
    ('output <PATH>', 'Outputs the data to a JSON file instead.'),
    ('center', 'Positions the text to be the center of the terminal window.')
]

argv = tuple(map(str.lower, argv[1:]))

class CLI(Client):
    __slots__ = tuple()

    def __init__(self):
        pass
    
    @property
    def client(self) -> Optional[Client]:
        global client_
        return client_

    @property
    def halo(self) -> "Halo":
        global halo_
        return halo_
    
    @property
    def location(self) -> Optional[str]:
        value = ''
        for arg in argv:
            if arg[0] == '-': return value[1:]
            value += ' ' + arg
        return value[1:]

    async def throw_err(self, message: str) -> None:
        self.halo.stop()
        print(f"\033[91merror: {message}\033[0m")
        if self.client:
            await self.client.close()
        exit(1)

    def get_arg(self, value: str) -> tuple:
        full = f'-{value[0]}' if f'-{value[0]}' in argv else f'--{value}'
        if full not in argv:
            return False, ''

        val = ''
        for part in argv[argv.index(full) + 1:]:
            if part[0] == '-':
                return True, val[1:]
            val += ' ' + part
        return True, val[1:]

    def display_help(self) -> None:
        global flags

        flags_str = ''
        for flag, description in flags:
            flag_name = flag.split(' ')[0]
            flags_str += f'\n  -\033[1m{flag_name[0]}\033[0m | --\033[1m{flag_name}\033[0m -> \033[1m{description}\033[0m\n  usage: \033[1m-{flag_name[0]} {flag[len(flag_name) + 1:]}\033[0m\n'

        self.halo.stop()
        print(f"\033[0;0Husage: \033[1m'weather <location> [...flags]'\033[0m\nflags:{flags_str[:-1]}")

    async def run(self) -> None:
        need_help, _ = self.get_arg('help')
        if need_help:
            return self.display_help()
        
        if not self.location:
            await self.throw_err("please add a location. run 'weather -h' for details.")

        _, unit = self.get_arg('unit')
        if unit and unit not in ('c', 'f', 'metric', 'imperial'):
            await self.throw_err("invalid unit. a unit must be 'c', 'f', 'metric', or 'imperial'.")
        
        unit = ('c' if (unit or 'metric') == 'metric' else 'f' if unit != 'c' else 'c').upper()
        _, locale = self.get_arg('locale')

        global client_
        client_ = Client(format=unit, locale=locale or "en-US")
        
        try:
            self.halo.text = f"requesting weather forecast for '{self.location}'..."
            response = await self.client.find(self.location)
        except HTTPException as http_err:
            await self.throw_err(f"error: {http_err.message.lower()}")

        self.halo.text = "formatting..."

        from datetime import datetime
        from re import findall
        
        current = response.current
        obj = {
            "location": response.location_name,
            "coordinates": f"{response.latitude}, {response.longitude}",
            "provider": response.provider,
            "temperature": f"{current.temperature} °{unit}",
            "feels like": f"{current.feels_like} °{unit}",
            "sky text": current.sky_text,
            "humidity": f"{current.humidity}%",
            "wind speeds": f"{current._get('@windspeed') or 'unknown'}",
            "date": datetime.strftime(current.date, f"%A, %d %B %Y at %H:%M:%S (UTC{'+' if response.timezone_offset > -1 else ''}{response.timezone_offset})"),
            "url": response.url
        }

        forecast_objects = {}
        forecast_objects_last_key = None
        for forecast in response.forecasts:
            formatted_date = datetime.strftime(forecast.date, f"%A, %d %B %Y at %H:%M:%S (UTC{'+' if response.timezone_offset > -1 else ''}{response.timezone_offset})")
            forecast_objects_last_key = formatted_date
            forecast_objects[formatted_date] = {
                "lowest temperature": f"{forecast.low} °{unit}",
                "highest temperature": f"{forecast.high} °{unit}",
                "sky text": forecast.sky_text,
                "precipitation": f"{forecast.precip}%" if forecast.precip else "unknown"
            }

        _, json_path = self.get_arg('output')
        if json_path:
            from json import dumps
            from os.path import isfile
            from re import search

            obj['forecasts'] = forecast_objects
            if not search(r'.json[/\\]?$', json_path):
                json_path += ".json"
            
            try:
                f = open(json_path, 'w+')
                f.write(dumps(obj).replace('\\u00b0', '°')) # why
                f.close()
            except Exception as err:
                await self.throw_err(err)
            self.halo.stop()
            print(f"\x1B[1m\x1B[32msuccess: saved json output in '{json_path}'.\x1B[39m\x1B[22m")
            return await self.client.close()

        centered_text, _ = self.get_arg('center')
        if centered_text:
            from os import get_terminal_size

        max_key_length = get_terminal_size().columns // 2 if centered_text else (max([
            *map(len, obj.keys()),
            *map(len, forecast_objects[forecast_objects_last_key].keys())
        ]) + 5)

        string = '\n'
        for key, value in obj.items():
            string += f"\x1B[90m{' ' * (max_key_length - len(key))}{key}\x1B[39m: \033[1m{value}\033[0m\n"

        for match in findall(r'https?://\S+', string):
            string = string.replace(match, f'\x1B[36m{match}\x1B[0m')
        
        if response.alert_message:
            string = f"\x1B[33m{' ' * (max_key_length - 5)}alert: {response.alert_message}\x1B[39m\n" + string

        for key, value in forecast_objects.items():
            if centered_text:
                string += f"\n{' ' * (max_key_length - ((len(key) + 2) // 2))}[{key}]\n"
            else:
                string += f"\n{' ' * max_key_length}  [{key}]\n"
            
            for k, v in value.items():
                string += f"\x1B[90m{' ' * (max_key_length - len(k))}{k}\x1B[39m: \033[1m{v}\033[0m\n"

        self.halo.stop()
        print(string[:-1])
        await self.client.close()

def main():
    cli = CLI()
    get_event_loop().run_until_complete(cli.run())

if __name__ == "__main__":
    main()


from datetime import datetime
from typing import List
from re import match

class BaseResponse:
    __slots__ = ('_get',)

    def __init__(self, response: dict):
        self._get = response.get
    
    def __getitem__(self, key: str):
        if (res := self._get(key)):
            return res
        raise KeyError(key)

class WeatherForecast(BaseResponse):
    def __init__(self, forecast_obj: dict):
        super().__init__(forecast_obj)

    def __repr__(self) -> str:
        return f"<WeatherForecast date={self.date!r} temperature={self.temperature} sky_text={self.sky_text} precip={self.precip}>"

    @property
    def low(self) -> int:
        """ Returns the lowest predicted temperature for the forecast. """
        if (temp := self._get('@low')):
            return int(temp)

    @property
    def high(self) -> int:
        """ Returns the highest predicted temperature for the forecast. """
        if (temp := self._get('@high')):
            return int(temp)

    @property
    def temperature(self) -> int:
        """ Returns the mean of the lowest and highest prediction. """
        return ((self.low or 0) + (self.high or 0)) // 2

    @property
    def sky_code_day(self) -> int:
        """ Returns the sky code day for the weather forecast. """
        if (code := self._get('@skycodeday')):
            return int(code)

    @property
    def sky_text(self) -> str:
        """ Returns the sky text day string. """
        return self._get('@skytextday')

    @property
    def date(self) -> "datetime":
        """ Returns the weather forecast date. """
        try:
            return datetime(*tuple(map(int, self._get('@date').split('-'))))
        except:
            pass

    @property
    def day(self) -> str:
        """ Returns the day the where the forecast was predicted to happen. """
        return self._get('@day')
    
    @property
    def short_day(self) -> str:
        """ Returns the day the where the forecast was predicted to happen. (short version) """
        return self._get('@shortday')

    @property
    def precip(self) -> int:
        """ Returns the precipitation value. """
        if (val := self._get('@precip')):
            return int(val)

class CurrentForecast(BaseResponse):
    def __init__(self, forecast_obj: dict):
        super().__init__(forecast_obj)
    
    def __repr__(self) -> str:
        return f"<CurrentForecast date={self.date!r} temperature={self.temperature} sky_text={self.sky_text} humidity={self.humidity} wind_speed={self.wind_speed}>"

    @property
    def temperature(self) -> int:
        """ Returns the current weather temperature. """
        if (temp := self._get('@temperature')):
            return int(temp)

    @property
    def sky_code(self) -> int:
        """ Returns the sky code for the current weather. """
        if (code := self._get('@skycode')):
            return int(code)

    @property
    def sky_text(self) -> str:
        """ Returns the sky text for the current weather. """
        return self._get('@skytext')

    @property
    def date(self) -> "datetime":
        """ Returns the observation date for the current weather. """
        try:
            args = tuple(map(int, self._get('@date').split('-'))) + tuple(map(int, self._get('@observationtime').split(':')))
            return datetime(*args)
        except:
            pass

    @property
    def observation_point(self) -> str:
        """ Returns the observation point for the current weather. """
        return self._get('@observationpoint')

    @property
    def feels_like(self) -> int:
        """ Returns the temperature of what it feels like. """
        if (temp := self._get('@feelslike')):
            return int(temp)
    
    @property
    def humidity(self) -> int:
        """ Returns the humidity level. """
        if (temp := self._get('@humidity')):
            return int(temp)
    
    @property
    def wind_display(self) -> str:
        """ Returns the wind display string. """
        return self._get('@winddisplay')

    @property
    def day(self) -> str:
        """ Returns the day the current weather was recorded. """
        return self._get('@day')
    
    @property
    def short_day(self) -> str:
        """ Returns the day the current weather was recorded. (short version) """
        return self._get('@shortday')
    
    @property
    def wind_speed(self) -> int:
        """ Returns the wind speed """
        try:
            return int(match(r'(\d+) (\w+)', self._get('@windspeed') or '')[1])
        except:
            pass

class Weather(BaseResponse):
    def __init__(self, response: dict):
        super().__init__(response)
    
    def __repr__(self) -> str:
        return f"<Weather latitude={self.latitude} longitude={self.longitude} location_name={self.location_name} url={self.url}>"

    @property
    def current(self) -> "CurrentForecast":
        """ Returns the CurrentForecast object for the weather object. """
        if (current := self._get('current')):
            return CurrentForecast(current)

    @property
    def forecasts(self) -> List[WeatherForecast]:
        """ Returns a list of forecasts for the weather object. """
        res = []
        for forecast in self._get('forecast', []):
            res.append(WeatherForecast(forecast))
        return res

    @property
    def location_code(self) -> str:
        """ Returns the weather location code. """
        return self._get('@weatherlocationcode')
    
    @property
    def location_name(self) -> str:
        """ Returns the weather location name. """
        return self._get('@weatherlocationname')
    
    @property
    def url(self) -> str:
        """ Returns the URL for the weather object. """
        return self._get('@url')
    
    @property
    def degree_type(self) -> str:
        """ Returns the format of the degree type. """
        return self._get('@degreetype')
    
    @property
    def provider(self) -> str:
        """ Returns the source provider for the weather forecast. """
        return self._get('@provider')
    
    @property
    def latitude(self) -> float:
        """ Returns the latitude of the weather's location. """
        if (lat := self._get('@lat')):
            return float(lat)
    
    @property
    def longitude(self) -> float:
        """ Returns the longitude of the weather's location. """
        if (long := self._get('@long')):
            return float(long)
    
    @property
    def timezone_offset(self) -> int:
        """ Returns the timezone offset. """
        if (offset := self._get('@timezone')):
            return int(offset)
    
    @property
    def alert_message(self) -> str:
        """ Returns the alert message if available. """
        return self._get('@alert') or None
    
    @property
    def entity_id(self) -> int:
        """ Returns the entity ID for the weather forecast. """
        if (id := self._get('@entityid')):
            return int(id)
    
    @property
    def provider_attribution(self) -> str:
        """ Returns the provider's attribution URL. """
        return self._get('@attribution')


from gc import collect

class Cache(dict):
    __slots__ = ('max_size',)

    def __repr__(self) -> str:
        return f"<Cache size={len(self.keys())} max_size={self.max_size}>"

    def __init__(self, max_size: int = 15) -> dict:
        self.max_size = max_size if isinstance(max_size, int) and (0 < max_size <= 100) else 15
        super().__init__()

    def __setitem__(self, key, value) -> None:
        if len(self.keys()) == self.max_size:
            dict.pop(self, list(self.keys())[0])
        
        dict.__setitem__(self, key, value)
    
    def clear(self) -> None:
        dict.clear(self)
        collect()


from setuptools import setup
setup(
  name='python-weather',
  packages=['python_weather'],
  version='0.3.3',
  license='MIT',
  description='A free and asynchronous Weather API Wrapper.',
  long_description=open('README.md', 'r', encoding='utf-8').read(),
  long_description_content_type='text/markdown',
  author='vierofernando',
  author_email='vierofernando9@gmail.com',
  url='https://github.com/vierofernando/python-weather',
  download_url='https://github.com/vierofernando/python-weather/archive/0.3.3.tar.gz',
  keywords=['Weather', 'API', 'Weather API', 'API Wrapper', 'Weather CLI', 'CLI'],
  install_requires=[
    'aiohttp>=3.7.4',
    'xmltodict',
    'halo'
  ],
  entry_points={
    'console_scripts': [
      'weather = python_weather.cli:main'
    ]
  },
  classifiers=[
    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Programming Language :: Python :: 3.9'
  ],
  python_requires='>=3.7',
)